/*******************************************************************************
    Copyright (c) 2013 NVIDIA Corporation

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to
    deal in the Software without restriction, including without limitation the
    rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
    sell copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be
        included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
    THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.
*******************************************************************************/

#if defined(DEBUG)

#include "nv_uvm_interface.h"
#include "uvm_gpu_ops_tests.h"
#include "uvm_page_migration.h"
#include "uvm_common.h"
#include "uvm_linux.h"
#include "uvm_lite_region_tracking.h"
#include "uvm_lite.h"
#include "uvm_channel_mgmt_common.h"

void gpuOpsSampleTest(UvmGpuUuid  * pUuidStruct)
{
    UVM_DBG_PRINT_UUID("Entering", pUuidStruct);
    return;
}

void pageMigrationTest(UvmGpuUuid  * pUuidStruct)
{
    RM_STATUS status;

    // arbitrarily choose that our region size will be 16KB
    const unsigned REGION_SIZE = 0x4000;
    const unsigned SEMA_SIZE =   4*1024;
    uvmGpuSessionHandle hSession;
    uvmGpuAddressSpaceHandle hVaSpace;
    uvmGpuChannelHandle hChannel;
    UvmGpuChannelPointers channelPointers;
    unsigned copyEngineClassNumber;
    uvmGpuCopyEngineHandle hCopyEngine;
    void * pbCpuPointer;
    UvmGpuPointer pbGpuPointer;
    UvmGpuPointer semaGpuPointer;
    void * semaCpuPointer;
    void * pbPutPointer;

    UvmGpuPointer fbRegionGpuPointer;
    void * fbRegionCpuPointer;
    void *pbEnd;
    UvmGpuPointer sysmemRegionGpuPointer;
    void * sysmemRegionCpuPointer;
    unsigned flags = 0;
    unsigned word;
    UvmCopyOps ceOps = {0};
    NvLength numMethods;
    unsigned semaVal;
    unsigned long long uuidMsb = 0;
    unsigned long long uuidLsb = 0;
    UvmGpuAllocInfo allocInfo;

    UVM_DBG_PRINT_UUID("Entering", pUuidStruct);
    memset(&allocInfo, 0, sizeof(allocInfo));

    //
    // allocate all resources necessary to utilize a CE:
    // 1) allocate cli/dev/subdev
    // 2) allocate channel from cli/dev VA space
    // 3) allocate OBJCE
    // 4) allocate PB
    //
    status = nvUvmInterfaceSessionCreate(&hSession);
    if (RM_OK != status)
    {
        UVM_ERR_PRINT_RM_STATUS("Could not create a session.", status);
        return;
    }

    memcpy(&uuidMsb, &pUuidStruct->uuid[0], (UVM_UUID_LEN >> 1));
    memcpy(&uuidLsb, &pUuidStruct->uuid[8], (UVM_UUID_LEN >> 1));


    status = nvUvmInterfaceAddressSpaceCreateMirrored(hSession,
                                                      uuidMsb,
                                                      uuidLsb,
                                                     &hVaSpace);
    if (RM_OK != status)
    {
        UVM_ERR_PRINT_RM_STATUS("Could not create an address space.", status);
        return;
    }

    status = nvUvmInterfaceChannelAllocate(hVaSpace, &hChannel,
                                          &channelPointers);
    if (RM_OK != status)
    {
        UVM_ERR_PRINT_RM_STATUS("Could not allocate a channel.", status);
        return;
    }

    status = nvUvmInterfaceCopyEngineAllocate(hChannel, 1,
                                             &copyEngineClassNumber,
                                             &hCopyEngine);
    if (RM_OK != status)
    {
        UVM_ERR_PRINT_RM_STATUS("Could not allocate OBJCE.", status);
        return;
    }

    // allocate semaphore page
    status = nvUvmInterfaceMemoryAllocSys(hVaSpace, SEMA_SIZE,
                                        &semaGpuPointer, NULL);
    if (RM_OK != status)
    {
        UVM_ERR_PRINT_RM_STATUS("Could not allocate GPU memory for PB.", status);
        return;
    }

    status = nvUvmInterfaceMemoryCpuMap(hVaSpace, semaGpuPointer,
                                        SEMA_SIZE, &semaCpuPointer);
    if (RM_OK != status)
    {
        UVM_ERR_PRINT_RM_STATUS("Could not map PB to CPU VA.", status);
        return;
    }

    // Push Buffer
    status = nvUvmInterfaceMemoryAllocSys(hVaSpace, REGION_SIZE,
                                         &pbGpuPointer, NULL);
    if (RM_OK != status)
    {
        UVM_ERR_PRINT_RM_STATUS("Could not allocate GPU memory for PB.", status);
        return;
    }
    // Map Push Buffer
    status = nvUvmInterfaceMemoryCpuMap(hVaSpace, pbGpuPointer,
                                        REGION_SIZE, &pbCpuPointer);
    if (RM_OK != status)
    {
        UVM_ERR_PRINT_RM_STATUS("Could not map PB to CPU VA.", status);
        return;
    }

    // allocate CPU accessible memory on FB that we can copy to/from for testing
    status = nvUvmInterfaceMemoryAllocFB(hVaSpace, REGION_SIZE,
                                        &fbRegionGpuPointer, &allocInfo);
    if (RM_OK != status)
    {
        UVM_ERR_PRINT_RM_STATUS("Could not allocate FB region for testing.",
                                status);
        return;
    }

    status = nvUvmInterfaceMemoryCpuMap(hVaSpace, fbRegionGpuPointer,
                                        REGION_SIZE, &fbRegionCpuPointer);
    if (RM_OK != status)
    {
        UVM_ERR_PRINT_RM_STATUS("Could not map GPU memory to CPU VA.",
                                status);
        return;
    }

    //
    // allocate CPU accessible memory on SYSMEM that we can copy to/from for
    // testing
    //
    status = nvUvmInterfaceMemoryAllocSys(hVaSpace, REGION_SIZE,
                                         &sysmemRegionGpuPointer, NULL);
    if (RM_OK != status)
    {
        UVM_ERR_PRINT_RM_STATUS("Could not allocate SYSMEM region for testing.",
                                status);
        return;
    }

    status = nvUvmInterfaceMemoryCpuMap(hVaSpace, sysmemRegionGpuPointer,
                                        REGION_SIZE,
                                       &sysmemRegionCpuPointer);
    if (RM_OK != status)
    {
        UVM_ERR_PRINT_RM_STATUS("Could not map SYSYMEM to CPU VA.",
                                status);
        return;
    }

    // setup CE Ops
    NvUvmHalInit(copyEngineClassNumber, 0x0000A06F, &ceOps);

    //setup to copy from SYSMEM to FB
    memset(fbRegionCpuPointer, 0xdeadbeef, REGION_SIZE);
    memset(sysmemRegionCpuPointer, 0xbad0cafe, REGION_SIZE);
    memset(semaCpuPointer, 0, SEMA_SIZE);


    // lets insert one sema and check for release
    pbPutPointer = pbCpuPointer;
    pbEnd = pbPutPointer + (REGION_SIZE / sizeof(unsigned));
    numMethods = ceOps.semaphoreRelease((unsigned **)&pbPutPointer,
                                        pbEnd,
                                        semaGpuPointer, 0xFACEFEED);

    // write the GP entry
    ceOps.writeGpEntry(channelPointers.gpFifoEntries, 0, pbGpuPointer,
                       numMethods);

    // launch the sema release
    NvUvmChannelWriteGpPut(channelPointers.GPPut, 1);

    // check if we get back the sempahore
    while(1)
        {
        semaVal = 0;
                semaVal = *(unsigned *)semaCpuPointer;
        UVM_DBG_PRINT("SemaVal:= 0x%x\n", semaVal);
        if(semaVal == 0xFACEFEED)
            break;
        }


    pbPutPointer += numMethods;
    pbGpuPointer += numMethods;
    pbEnd = pbPutPointer + (REGION_SIZE / sizeof(unsigned));

    // push methods to do a copy
    numMethods = ceOps.launchDma((unsigned **)&pbPutPointer,
                    pbEnd,
                    sysmemRegionGpuPointer, 2,
                    fbRegionGpuPointer, 0,
                    REGION_SIZE, flags);

    // write the GP entry
    ceOps.writeGpEntry(channelPointers.gpFifoEntries, 1, pbGpuPointer,
                       numMethods);

    // launch the copy
    NvUvmChannelWriteGpPut(channelPointers.GPPut, 2);


    // check that the pattern copied to FB is the same as the pattern in SYSMEM
    for (word = 0; word < REGION_SIZE / sizeof(unsigned); word ++)
    {
        UVM_INFO_PRINT("INFO: checking 0x%X == 0x%X ?\n",
                       ((unsigned*)fbRegionCpuPointer)[word],
                       ((unsigned*)sysmemRegionCpuPointer)[word]);
        if (((unsigned*)fbRegionCpuPointer)[word] !=
            ((unsigned*)sysmemRegionCpuPointer)[word])
        {
            UVM_DBG_PRINT_RL("ERROR: copy failed. FB region = 0x%X,"
                             " SYSMEM region = 0x%X\n",
                             ((unsigned*)fbRegionCpuPointer)[word],
                             ((unsigned*)sysmemRegionCpuPointer)[word]);

            nvUvmInterfaceSessionDestroy(hSession);
            return;
        }
    }
    nvUvmInterfaceChannelDestroy(hChannel);
    nvUvmInterfaceAddressSpaceDestroy(hVaSpace);
    nvUvmInterfaceSessionDestroy(hSession);

    UVM_DBG_PRINT("INFO: Passed\n");
}

//
// Test copying from sysmem to fb.
//
void channelMgmtApiBasicMigrationTest(UvmGpuUuid *pGpuUuidStruct)
{
    RM_STATUS status;

    // arbitrarily choose that our region size will be 16KB
    const unsigned REGION_SIZE = 0x4000;
    const unsigned SEMA_SIZE   = 0x1000;

    UvmChannelManager *channelManager = NULL;

    UvmRingbufferPool *ringbufferPool = NULL;
    UvmRingbuffer     *ringbuffer     = NULL;

    UvmChannelPool *channelPool = NULL;
    UvmChannel     *channel     = NULL;

    UvmGpuPointer fbRegionGpuPointer     = 0;
    void         *fbRegionCpuPointer     = NULL;
    UvmGpuPointer sysmemRegionGpuPointer = 0;
    void         *sysmemRegionCpuPointer = NULL;
    UvmGpuPointer semaGpuPointer         = 0;
    void         *semaCpuPointer         = NULL;

    UvmGpuPointer  pbGpuPointer = 0;
    void          *pbPutPointer = NULL;
    void          *pbEnd        = NULL;
    NvLength       numMethods   = 0;
    unsigned       word         = 0;
    UvmGpuAllocInfo allocInfo;

    UVM_DBG_PRINT_UUID("Entering",
                       pGpuUuidStruct);

    memset(&allocInfo, 0, sizeof(allocInfo));
    status = uvm_initialize_channel_mgmt_api();
    if (RM_OK != status)
    {
        UVM_ERR_PRINT_RM_STATUS("could not initialize channel mgmt api.",
                                status);
        goto cleanup;
    }

    status = uvm_create_channel_manager(pGpuUuidStruct, &channelManager);
    if (RM_OK != status)
    {
        UVM_ERR_PRINT_RM_STATUS("Could not allocate UvmManager.",
                                status);
        goto cleanup;
    }

    channelPool    = &channelManager->channelPool;
    ringbufferPool = &channelManager->ringbufferPool;

    // get the first ringbuffer
    ringbuffer = list_entry(ringbufferPool->ringbufferFreeListHead.next,
                            UvmRingbuffer, ringbufferList);
    // get the first channel
    channel = list_entry(channelPool->channelListHead.next,
                         UvmChannel, channelList);

    // allocate CPU accessible memory on FB that we can copy to/from for testing
    status = nvUvmInterfaceMemoryAllocFB(channelPool->hVaSpace, REGION_SIZE,
                                         &fbRegionGpuPointer, &allocInfo);
    if (RM_OK != status)
    {
        UVM_ERR_PRINT_RM_STATUS("Could not allocate FB region for testing.",
                                status);
        goto cleanup;
    }

    status = nvUvmInterfaceMemoryCpuMap(channelPool->hVaSpace,
                                        fbRegionGpuPointer,
                                        REGION_SIZE,
                                       &fbRegionCpuPointer);
    if (RM_OK != status)
    {
        UVM_ERR_PRINT_RM_STATUS("Could not map GPU memory to CPU VA.", status);
        goto cleanup;
    }

    //
    // allocate CPU accessible memory on SYSMEM that we can copy to/from for
    // testing
    //
    status = nvUvmInterfaceMemoryAllocSys(channelPool->hVaSpace, REGION_SIZE,
                                         &sysmemRegionGpuPointer, NULL);
    if (RM_OK != status)
    {
        UVM_ERR_PRINT_RM_STATUS("Could not allocate SYSMEM region for testing",
                                status);
        goto cleanup;
    }

    status = nvUvmInterfaceMemoryCpuMap(channelPool->hVaSpace,
                                        sysmemRegionGpuPointer,
                                        REGION_SIZE,
                                       &sysmemRegionCpuPointer);
    if (RM_OK != status)
    {
        UVM_ERR_PRINT_RM_STATUS("Could not map SYSYMEM to CPU VA.", status);
        goto cleanup;
    }

    // allocate semaphore page
    status = nvUvmInterfaceMemoryAllocSys(channelPool->hVaSpace,
                                          SEMA_SIZE,
                                         &semaGpuPointer, NULL);
    if (RM_OK != status)
    {
        UVM_ERR_PRINT_RM_STATUS("Could not allocate GPU memory for sema.",
                                status);
        goto cleanup;
    }

    status = nvUvmInterfaceMemoryCpuMap(channelPool->hVaSpace,
                                        semaGpuPointer,
                                        SEMA_SIZE, &semaCpuPointer);
    if (RM_OK != status)
    {
        UVM_ERR_PRINT_RM_STATUS("Could not map sema to CPU VA.", status);
        goto cleanup;
    }

    //setup to copy from SYSMEM to FB
    memset(fbRegionCpuPointer, 0xFF, REGION_SIZE);
    memset(sysmemRegionCpuPointer, 0x1, REGION_SIZE);
    memset(semaCpuPointer, 0, SEMA_SIZE);

    UVM_INFO_PRINT("\n"
                   "INFO: fb     cpu: 0x%X\n"
                   "INFO: sysmem cpu: 0x%X\n"
                   "INFO: sema   cpu: 0x%X\n",
                   *((unsigned*)fbRegionCpuPointer),
                   *((unsigned*)sysmemRegionCpuPointer),
                   *((unsigned*)semaCpuPointer));

    pbPutPointer = ringbuffer->cpuBegin;
    pbGpuPointer = ringbuffer->gpuBegin;
    pbEnd = ringbuffer->cpuEnd;

    //
    // test migration
    //

    // push methods to pb to do a copy
    numMethods = channel->ceOps.launchDma(
            (unsigned **)&pbPutPointer,
            (unsigned *)pbEnd,
            sysmemRegionGpuPointer, NV_UVM_COPY_SRC_LOCATION_SYSMEM,
            fbRegionGpuPointer, NV_UVM_COPY_DST_LOCATION_FB,
            REGION_SIZE,
            NV_UVM_COPY_SRC_TYPE_VIRTUAL | NV_UVM_COPY_DST_TYPE_VIRTUAL);

    // write the GP entry
    channel->ceOps.writeGpEntry(
            channel->channelInfo.gpFifoEntries,
            channel->trackingInfo.curGpFifoOffset,
            pbGpuPointer,
            numMethods);

    // increment current GPFIFO offset and launch copy
    if (++channel->trackingInfo.curGpFifoOffset >=
        channel->channelInfo.numGpFifoEntries)
    {
        channel->trackingInfo.curGpFifoOffset = 0;
    }
    NvUvmChannelWriteGpPut(channel->channelInfo.GPPut,
                           channel->trackingInfo.curGpFifoOffset);

    // increment GPU pointer
    pbGpuPointer += numMethods;

    // insert one sema and check for release
    numMethods = channel->ceOps.semaphoreRelease(
            (unsigned **)&pbPutPointer,
            pbEnd,
            semaGpuPointer, 0xFACEFEED);

    // write the GP entry
    channel->ceOps.writeGpEntry(
            channel->channelInfo.gpFifoEntries,
            channel->trackingInfo.curGpFifoOffset,
            pbGpuPointer,
            numMethods);

    // increment current GPFIFO offset and launch sema release
    if (++channel->trackingInfo.curGpFifoOffset >=
        channel->channelInfo.numGpFifoEntries)
    {
        channel->trackingInfo.curGpFifoOffset = 0;
    }
    NvUvmChannelWriteGpPut(channel->channelInfo.GPPut,
                           channel->trackingInfo.curGpFifoOffset);

    // check if we get back the sempahore
    {
        int count = 0;
        while(1)
        {
            unsigned semaVal = 0;
            semaVal = *(unsigned *)semaCpuPointer;
            UVM_DBG_PRINT("SemaVal:= 0x%x\n", semaVal);
            if (semaVal == 0xFACEFEED)
                break;

            if (++count > 1000)
            {
                UVM_ERR_PRINT("ERROR: sema release timed out.\n");
                break;
            }
        }
    }

    // check that the pattern copied to FB is the same as the pattern in SYSMEM
    for (word = 0; word < REGION_SIZE / sizeof(unsigned); word ++)
    {
        if (((unsigned*)fbRegionCpuPointer)[word] !=
            ((unsigned*)sysmemRegionCpuPointer)[word])
        {
            UVM_ERR_PRINT("ERROR: copy failed. FB region = 0x%X,"
                          " SYSMEM region = 0x%X\n",
                          ((unsigned*)fbRegionCpuPointer)[word],
                          ((unsigned*)sysmemRegionCpuPointer)[word]);
            break;
        }
    }

cleanup:
    if (semaCpuPointer)
        nvUvmInterfaceMemoryCpuUnMap(channelPool->hVaSpace,
                                     semaCpuPointer);

    if (semaGpuPointer)
        nvUvmInterfaceMemoryFree(channelPool->hVaSpace,
                                 semaGpuPointer);

    if (sysmemRegionCpuPointer)
        nvUvmInterfaceMemoryCpuUnMap(channelPool->hVaSpace,
                                     sysmemRegionCpuPointer);

    if (sysmemRegionGpuPointer)
        nvUvmInterfaceMemoryFree(channelPool->hVaSpace,
                                 sysmemRegionGpuPointer);

    if (fbRegionCpuPointer)
        nvUvmInterfaceMemoryCpuUnMap(channelPool->hVaSpace,
                                     fbRegionCpuPointer);

    if (fbRegionGpuPointer)
        nvUvmInterfaceMemoryFree(channelPool->hVaSpace,
                                 fbRegionGpuPointer);

    if (channelManager)
    {
        uvm_destroy_channel_manager(channelManager);
    }

    uvm_deinitialize_channel_mgmt_api();
}

//
// Simple Pushbuffer Sanity Test
//
//   1. Get a pushbuffer.
//  
//   2. Push commands to copy first half of FB to SYSMEM.
//  
//   3. Submit and get tracker t1.
//  
//   4. Get a second pushbuffer.
//  
//   5. Push commands to copy second half of FB to SYSMEM.
//  
//   6. Submit: depend on t1, and get back tracker t2.
//  
//   7. Wait on t2 on CPU-side.
//
void channelMgmtApiPushbufferSimpleSanityTest(UvmGpuUuid *pGpuUuidStruct)
{
    RM_STATUS status;
    NvBool ret;
    unsigned index;

    // arbitrarily choose that our region size will be 32 * 4KB
    const unsigned REGION_SIZE = 0x1000 * 32;

    UvmTracker tracker1;
    UvmTracker tracker2;

    UvmPushbuffer     *pushbuffer     = NULL;
    UvmChannelManager *channelManager = NULL;

    UvmChannelPool *channelPool = NULL;
    UvmChannel     *channel     = NULL;

    uvmGpuAddressSpaceHandle hVaSpace;

    UvmGpuPointer fbRegionGpuPointer     = 0;
    void         *fbRegionCpuPointer     = NULL;
    UvmGpuPointer sysmemRegionGpuPointer = 0;
    void         *sysmemRegionCpuPointer = NULL;
    UvmGpuAllocInfo allocInfo;

    unsigned word = 0;

    UVM_DBG_PRINT_UUID("Entering",
                       pGpuUuidStruct);

    memset(&allocInfo, 0, sizeof(allocInfo));
    status = uvm_initialize_channel_mgmt_api();
    if (RM_OK != status)
    {
        UVM_ERR_PRINT_RM_STATUS("could not initialize channel mgmt api.",
                                status);
        goto cleanup;
    }

    status = uvm_create_channel_manager(pGpuUuidStruct, &channelManager);
    if (RM_OK != status)
    {
        UVM_ERR_PRINT_RM_STATUS("Could not allocate UvmManager.",
                                status);
        goto cleanup;
    }

    channelPool = &channelManager->channelPool;
    hVaSpace = channelPool->hVaSpace;

    // allocate CPU accessible memory on FB that we can copy to/from for testing
    status = nvUvmInterfaceMemoryAllocFB(hVaSpace, REGION_SIZE,
                                         &fbRegionGpuPointer, &allocInfo);
    if (RM_OK != status)
    {
        UVM_ERR_PRINT_RM_STATUS("Could not allocate FB region for testing.",
                                status);
        goto cleanup;
    }

    status = nvUvmInterfaceMemoryCpuMap(hVaSpace,
                                        fbRegionGpuPointer,
                                        REGION_SIZE,
                                       &fbRegionCpuPointer);
    if (RM_OK != status)
    {
        UVM_ERR_PRINT_RM_STATUS("Could not map GPU memory to CPU VA.", status);
        goto cleanup;
    }

    //
    // allocate CPU accessible memory on SYSMEM that we can copy to/from for
    // testing
    //
    status = nvUvmInterfaceMemoryAllocSys(hVaSpace, REGION_SIZE,
                                         &sysmemRegionGpuPointer, NULL);
    if (RM_OK != status)
    {
        UVM_ERR_PRINT_RM_STATUS("Could not allocate SYSMEM region for testing",
                                status);
        goto cleanup;
    }

    status = nvUvmInterfaceMemoryCpuMap(hVaSpace,
                                        sysmemRegionGpuPointer,
                                        REGION_SIZE,
                                       &sysmemRegionCpuPointer);
    if (RM_OK != status)
    {
        UVM_ERR_PRINT_RM_STATUS("Could not map SYSYMEM to CPU VA.", status);
        goto cleanup;
    }

    //
    // test migration
    //
    for (index = 0; index < 8; ++index)
    {
        //setup to copy from SYSMEM to FB
        memset(fbRegionCpuPointer, 0xFF, REGION_SIZE);
        memset(sysmemRegionCpuPointer, 0x1, REGION_SIZE);

        UVM_INFO_PRINT("%dth loop: "
                "fb cpu: 0x%X "
                "sysmem cpu: 0x%X\n",
                index,
                *((unsigned*)fbRegionCpuPointer),
                *((unsigned*)sysmemRegionCpuPointer));

        // get 1st pushbuffer
        status = uvm_get_pushbuffer(channelManager, &pushbuffer);
        if (RM_OK != status)
        {
            UVM_ERR_PRINT_RM_STATUS("Could not get 1st pushbuffer.",
                    status);
            goto cleanup;
        }

        // get channel associated with 1st pushbuffer
        channel = pushbuffer->channel;

        // push methods to pb to do a copy over first half of the range
        UVM_PUSH_METHOD(
                ret, pushbuffer, channel->ceOps.launchDma,
                sysmemRegionGpuPointer, NV_UVM_COPY_SRC_LOCATION_SYSMEM,
                fbRegionGpuPointer, NV_UVM_COPY_DST_LOCATION_FB,
                REGION_SIZE / 2,
                NV_UVM_COPY_SRC_TYPE_VIRTUAL | NV_UVM_COPY_DST_TYPE_VIRTUAL);
        if (!ret)
        {
            status = RM_ERR_NO_MEMORY;
            UVM_ERR_PRINT_RM_STATUS("Could not push method to 1st pb.", status);
            goto cleanup;
        }

        // submit pushbuffer
        status = uvm_submit_pushbuffer(channelManager, pushbuffer,
                NULL, &tracker1);
        if (status != RM_OK)
        {
            UVM_ERR_PRINT_RM_STATUS("Could not submit 1st pushbuffer.", status);
            goto cleanup;
        }

        // get 2nd pushbuffer
        status = uvm_get_pushbuffer(channelManager, &pushbuffer);
        if (RM_OK != status)
        {
            UVM_ERR_PRINT_RM_STATUS("Could not get 2nd pushbuffer.",
                    status);
            goto cleanup;
        }

        // get channel associated with 2nd pushbuffer
        channel = pushbuffer->channel;

        // push methods to pb to do a copy over 2nd half of the range
        UVM_PUSH_METHOD(
                ret, pushbuffer, channel->ceOps.launchDma,
                sysmemRegionGpuPointer + REGION_SIZE / 2,
                NV_UVM_COPY_SRC_LOCATION_SYSMEM,
                fbRegionGpuPointer + REGION_SIZE / 2,
                NV_UVM_COPY_DST_LOCATION_FB,
                REGION_SIZE / 2,
                NV_UVM_COPY_SRC_TYPE_VIRTUAL | NV_UVM_COPY_DST_TYPE_VIRTUAL);
        if (!ret)
        {
            status = RM_ERR_NO_MEMORY;
            UVM_ERR_PRINT_RM_STATUS("Could not push method to 2nd pb.", status);
            goto cleanup;
        }

        // submit this pushbuffer that depends on tracker1
        status = uvm_submit_pushbuffer(channelManager, pushbuffer,
                &tracker1, &tracker2);
        if (status != RM_OK)
        {
            UVM_ERR_PRINT_RM_STATUS("Could not submit 2nd pushbuffer.", status);
            goto cleanup;
        }

        //
        // wait for tracker2 on cpu-side.
        // (which effectively also waits for tracker1)
        //
        uvm_wait_for_tracker(&tracker2);

        // check that the pattern copied to FB is the same as the pattern in SYSMEM
        for (word = 0; word < REGION_SIZE / sizeof(unsigned); word ++)
        {
            if (((unsigned*)fbRegionCpuPointer)[word] !=
                    ((unsigned*)sysmemRegionCpuPointer)[word])
            {
                UVM_ERR_PRINT("ERROR: copy failed. FB region = 0x%X,"
                        " SYSMEM region = 0x%X\n",
                        ((unsigned*)fbRegionCpuPointer)[word],
                        ((unsigned*)sysmemRegionCpuPointer)[word]);
                break;
            }
        }
    }

cleanup:
    if (sysmemRegionCpuPointer)
        nvUvmInterfaceMemoryCpuUnMap(hVaSpace,
                                     sysmemRegionCpuPointer);

    if (sysmemRegionGpuPointer)
        nvUvmInterfaceMemoryFree(hVaSpace,
                                 sysmemRegionGpuPointer);

    if (fbRegionCpuPointer)
        nvUvmInterfaceMemoryCpuUnMap(hVaSpace,
                                     fbRegionCpuPointer);

    if (fbRegionGpuPointer)
        nvUvmInterfaceMemoryFree(hVaSpace,
                                 fbRegionGpuPointer);

    if (pushbuffer && status != RM_OK)
        uvm_cancel_pushbuffer(channelManager, pushbuffer);

    if (channelManager)
    {
        uvm_destroy_channel_manager(channelManager);
    }

    uvm_deinitialize_channel_mgmt_api();
}

void _destroy_dummy_region(UvmCommitRecord * commit)
{
    const unsigned long long REGION_NUMBER = 0x1000;
    // The commit is supposed to be a fake commit
    // just check that the pointer contain the magick number
    if (commit > (UvmCommitRecord *) REGION_NUMBER)
        UVM_ERR_PRINT_RM_STATUS("Ask to destroy an invalid commit",
                                RM_ERR_INVALID_ARGUMENT);
}

//
// Simple Region Tracking Sanity Test
//
//   1. Create a region tracking tree
//
//   2. Add a regions in the tree
//
//   3. Check the regions
//
//   4. Destroy the region tracking tree
//
void regionTrackerSanityTest(void)
{
    const unsigned long long REGION_NUMBER = 0x1000;
    const unsigned long long REGION_SIZE = 0x1000;

    RM_STATUS status;
    struct vm_area_struct vma;
    UvmRegionTracker *region_tracker = NULL;
    unsigned long long region = 0;
    unsigned long long offset = 0;

    // Create a fake vma
    vma.vm_start = 0x0;
    vma.vm_end = (unsigned long)(REGION_NUMBER * REGION_SIZE);

    region_tracker = uvm_create_region_tracker(&vma);

    if (!region_tracker)
    {
        status = RM_ERR_NO_MEMORY;
        UVM_ERR_PRINT_RM_STATUS("Could not create a region tracking tree.",
                                status);
        goto cleanup;
    }

    for (region = 0; region < REGION_NUMBER; region++)
    {
      unsigned long long fakeRegionStart = region * REGION_SIZE;
      unsigned long long fakeRegionEnd = (region + 1) * REGION_SIZE;

      // Create a fake commit record pointer that is only an incremental
      // counter
      UvmCommitRecord * fakeCommitRecord = (UvmCommitRecord *) (region);

      uvm_track_region(region_tracker,
                       fakeRegionStart,
                       fakeRegionEnd, NULL,
                       fakeCommitRecord);
    }

    // Check that everything is correct
    for (region = 0; region < REGION_NUMBER; region++)
    {
        unsigned long long fakeRegionStart = region * REGION_SIZE;
        UvmCommitRecord * fakeCommitRecordExpected = (UvmCommitRecord *) (region);
        UvmCommitRecord *commitToCheck;
        for (offset = 0; offset < REGION_SIZE; offset++)
        {
            status = uvm_get_owner_from_address(region_tracker,
                                                fakeRegionStart + offset,
                                                &commitToCheck);
            if (RM_OK != status)
            {
                UVM_ERR_PRINT_RM_STATUS("Could not get information for the address.",
                                        status);
                goto cleanup;
            }
            // The fake commit record is supposed to be an incremental counter
            // Just check if the expected value of the counter is the one store
            // in the region tracking tree
            if (commitToCheck != fakeCommitRecordExpected)
            {
                status = RM_ERR_INVALID_ARGUMENT;
                UVM_ERR_PRINT_RM_STATUS("Invalid owner in the region tracking tree.",
                                        status);
            }
        }
    }

cleanup:
    uvm_destroy_region_tracker(region_tracker, _destroy_dummy_region);
}

#endif
